/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "bangablemenu.h"
#include "globals.h"
#include <macros.h>

BangableMenu::BangableMenu(const std::string& bang, const MenuSubPtr& sub, HWND parenthwnd, BangCommandEx* f)
	:Window(W_MODULE_NAME),
	m_menu(sub),
	m_bangcmd(bang),
	m_autobang(f != NULL)
{
	if (sub)
		create(sub->getEntry().c_str(), parenthwnd);
	else
		create(W_MODULE_NAME, parenthwnd);
	
	if (m_autobang)
		AddBangCommandEx(m_bangcmd.c_str(), f);
}

BangableMenu::BangableMenu(HWND parenthwnd)
	:Window(W_MODULE_NAME),
	m_autobang(false)
{
	create(W_MODULE_NAME, parenthwnd);
}

BangableMenu::~BangableMenu()
{
	if (m_autobang)
		RemoveBangCommand(m_bangcmd.c_str());
	destroyWindow();
}

void BangableMenu::onBang(const char *pszArguments)
{
	POINT pt;
	GetCursorPos(&pt);

	// Handle any params that come with the bang
	char sztoken1[MAX_RC_STRING];
	char sztoken2[MAX_RC_STRING];
	char* rgsztokens[] = { sztoken1, sztoken2 };

	int ntokens = LCTokenize(pszArguments, rgsztokens, 2, NULL);
	// There are actually two more possible args, however they cannot be implemented with menu

	// this huge comparison is because pszArguments will always return something
	// it should IMO return NULL, "", or have another arg telling you how long the args are
	// if there are no params, but it doesn't
	// even this will not catch every possiblity, but anything that is incorrect that gets past
	// this will have the menu display at 0,0 
	// this is still a problem here, a exception is thrown
	if (ntokens >= 2)
	{

		if (isdigit(sztoken1[0]) || sztoken1[0] == '-' || _strnicmp(sztoken1, "X=", 2) == 0)
		{
			int x, y;
			char* ptok1 = sztoken1;
			char* ptok2 = sztoken2;

			if (_strnicmp(sztoken1, "X=", 2) == 0)
			{
				x = atoi(sztoken1 + 2);
				ptok1 = sztoken1 +  2;
			}
			else
				x = atoi(sztoken1);

			if (_strnicmp(sztoken2, "y=", 2) == 0)
			{
				y = atoi(sztoken2 + 2);
				ptok2 = sztoken2 +  2;
			}
			else
				y = atoi(sztoken2);

			if (x != 0 && ptok1[0] != '0')
				pt.x = CONVERT_COORDINATE_X(x);
			if (y != 0 && ptok2[0] != '0')
				pt.y = CONVERT_COORDINATE_Y(y);
		}	
	}

	m_menu->displayMenu(pt.x, pt.y, hWnd); 
}

void BangableMenu::windowProc(Message& message)
{
	try
	{
		BEGIN_MESSAGEPROC
			MESSAGE(onEndSession,    WM_ENDSESSION)
			MESSAGE(onEndSession,    WM_QUERYENDSESSION)
			MESSAGE(onSysCommand,    WM_SYSCOMMAND)

			MESSAGE(onCommand,       WM_COMMAND)
			MESSAGE(onTimer,         WM_TIMER)
			MESSAGE(onMenuRButtonUp, WM_MENURBUTTONUP)
			MESSAGE(onEnterMenuLoop, WM_ENTERMENULOOP)
			MESSAGE(onExitMenuLoop,  WM_EXITMENULOOP)
			MESSAGE(onInitMenuPopup, WM_INITMENUPOPUP)
		END_MESSAGEPROC
	}
	catch (std::exception& e)
	{
		LSLog(LOG_DEBUG, MODULE_NAME, e.what());
	}
	catch (...)
	{
		LSLog(LOG_DEBUG, MODULE_NAME, "Unknown Exception Caught");
	}
}

void BangableMenu::onEndSession(Message& message)
{
	message.lResult = SendMessage(hParent, message.uMsg, message.wParam, message.lParam);
}

void BangableMenu::onSysCommand(Message& message)
{
	if (message.wParam == SC_CLOSE)
		PostMessage(hParent, WM_KEYDOWN, LM_SHUTDOWN, 0);
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}

void BangableMenu::onCommand(Message& message)
{
	if (message.wParamHi == 0)	// Is from a menu
		if (MenuEntryPtr meptr = m_menu->getItem((unsigned int)message.wParamLo))
			meptr->execute(hWnd);
}

void BangableMenu::onTimer(Message& message)
{
	KillTimer(hWnd, 1);		// Could Error, Not Checking
	m_menu->endDisplay();
}

void BangableMenu::onMenuRButtonUp(Message& message)
{
	MENUITEMINFO info;
	info.cbSize = sizeof(info);
	info.fMask = MIIM_FTYPE | MIIM_ID | MIIM_STATE | MIIM_SUBMENU;
	if (!GetMenuItemInfo((HMENU)message.lParam, (UINT)message.wParam, true, &info))
	{
		LSLog(LOG_DEBUG, MODULE_NAME, ("GetMenuItemInfo Failed. " + getWinErrorMsg()).c_str());
		return;
	}

	if (!(info.fType & MFT_SEPARATOR) && !(info.fState & MFS_DISABLED))
	{
		MenuEntryPtr selectedItem;
		 if (info.hSubMenu == NULL)
			selectedItem = m_menu->getItem(info.wID);
		else
			selectedItem = m_menu->getSub(info.hSubMenu);
		
		// Verify that we really found the item
		if (selectedItem)
		{
			if (selectedItem->displayContextMenu(hWnd))
				EndMenu();
		}
	}
}

void BangableMenu::onEnterMenuLoop(Message& message)
{
	// Kill the timer so it doesn't destroy the dynamic menu's while in use
	KillTimer(hWnd, 1);
	
	// Get Focus so that keyboard input will work
	SwitchToThisWindow(hWnd, FALSE);
}

void BangableMenu::onExitMenuLoop(Message& message)
{
	if (!SetTimer(hWnd, 1, 250, (TIMERPROC) NULL))
		LSLog(LOG_DEBUG, MODULE_NAME, ("SetTimer Failed. " + getWinErrorMsg()).c_str());
}

void BangableMenu::onInitMenuPopup(Message& message)
{
	if (MenuSubPtr msptr = m_menu->getSub((HMENU)message.wParam))
		msptr->preDisplay(hWnd);
}