/*
* Copyright 2003, Nathan E. Moore
*
* This file is part of Menu.
*
* Menu is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* Menu is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Menu; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#pragma once
#include <stack>
#include <vector>
#include <string>
#include <algorithm>
#include <boost\shared_ptr.hpp>
#include <windows.h>

class MenuEntry;
typedef boost::shared_ptr<MenuEntry> MenuEntryPtr;

bool nocase(wchar_t x, wchar_t y);

class MenuEntry
{
private:
	// Never mess with the id!
	const unsigned int m_id;

	// The text you see when the menu is displayed
	std::wstring m_entry;

	// May change to a random number
	static unsigned int idnext;
	static std::stack<unsigned int, std::vector<unsigned int> >idstack;
	// First id returned is NULL_ID + 1, NULL_ID is used as the id for all dead items
	static unsigned int getNextId();

protected:
	// Same as the id used for separators 
	static const unsigned int NULL_ID = 0;

	explicit MenuEntry(const std::wstring& txt)
		:m_id(getNextId()),
		m_entry(txt) 
	{	}

	// after addition to a MenuSub this function will not change the displayed text
	// (although it should and will in a later release),
	// this function is for use in a constructor
	// this set of getter / setter methods are need to allow for subclasses to receive
	// notification when the Entry text changes.
	virtual void setEntry(std::wstring& txt)
	{
		m_entry = txt;
	}

public:
	virtual ~MenuEntry()
	{
		// return the id so that it can be used again
		idstack.push(m_id);
	}
	
	const unsigned int getId() const
	{
		return m_id;
	}

	// Not sure if this really should return a reference, but that is what it does
	const std::wstring& getEntry() const
	{
		return m_entry;
	}  

	// Executes Item 
	virtual void execute(HWND hwnd) = 0;

	// Displays Context Menu for the item
	virtual bool displayContextMenu(HWND hwnd) = 0;

	static bool order(MenuEntryPtr lhs, MenuEntryPtr rhs)
	{
		std::wstring& lhsstr = lhs->m_entry;
		std::wstring& rhsstr = rhs->m_entry;
		return std::lexicographical_compare(lhsstr.begin(), lhsstr.end(), rhsstr.begin(), rhsstr.end(), nocase); 
	}
};