/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "menufolder.h"
#include "menuitem.h"
#include "globals.h"
#include <algorithm>
#include <boost/bind.hpp>
#include <stdexcept>
#include <windows.h>
#include <shlwapi.h>
#include <lsapi.h>

void MenuFolder::PathObject::validate()
{
	// 1) Determine if a pattern is part of this->path
	// 2) if so seperate it out into this->file
	// 3) otherwise backslashify this->path

	size_t location = m_path.find_last_of(L'*');
	if (location == std::wstring::npos)
	{
		location = m_path.find_last_of(L'?');
		if (location == std::wstring::npos)
		{
			m_file = L"*";
			
			// Do PathAddBackslash without the messy conversion to wchar_t
			if(*m_path.rbegin() != L'\\')
				m_path.append(L"\\");
		}
	}
	else	// There is a pattern
	{
		location = m_path.find_last_of(L'\\');
		if (location != std::wstring::npos)	// Should be redundent
		{
			location++;
			m_file = m_path.substr(location, m_path.length() - location);
			m_path = m_path.substr(0, location);
		}
	}
}

MenuFolder::MenuFolder(const std::wstring& pathname, bool isdynamic)
	:MenuAuto(std::wstring(), isdynamic)
{
	create(pathname);

	// find the last two '/' in the path
	// cut out the text in between
	// or if i has only one '/' then just use the path
	// use said text to setEntry

	PathVector::reference totitle = m_pathvect.front();
	std::wstring::size_type lastslash = totitle.m_path.find_last_of(L'\\');
	std::wstring::size_type slash = totitle.m_path.find_last_of(L'\\', lastslash - 1);
	if (slash == std::wstring::npos)
		setEntry(totitle.m_path);
	else
		setEntry(totitle.m_path.substr(slash + 1, lastslash - slash - 1));
}

void MenuFolder::create(const std::wstring& path)
{
	StrVector strvect;
	strvect.push_back(path);
	validatePath(strvect);	// recursive Function
	for (StrVector::iterator i = strvect.begin(); i != strvect.end(); ++i)
		m_pathvect.push_back(PathObject(*i));
}

void MenuFolder::parse()
{
	if (m_pathvect.empty())
		return;

	EntryVector items;
	FolderVector subs;

	for (PathVector::iterator i = m_pathvect.begin(); i != m_pathvect.end(); i++)
	{
		parseFolder(*i, items, subs);
		std::sort(subs.begin(), subs.end(), MenuEntry::order);
	}
	//for_each(subs.begin(), subs.end(), boost::bind(&MenuFolder::addSubMenu, this, _1, false));
	for (FolderVector::iterator i = subs.begin(); i != subs.end(); ++i)
		addSubMenu(boost::static_pointer_cast<MenuSub>(*i));
	
	std::sort(items.begin(), items.end(), MenuEntry::order);
	std::for_each(items.begin(), items.end(), boost::bind(&MenuFolder::addItem, this, _1, false));
}

void MenuFolder::parseFolder(PathObject& parsepath, EntryVector& itemlist, FolderVector& sublist)
{
	HANDLE hfind;
	WIN32_FIND_DATA FindFileData;

	if (getEntry() == L"emule")
		int jackass = 39*5/5-51;

	// The \\?\ allows for absurdly long paths to be used
	hfind = FindFirstFile((std::wstring(L"\\\\?\\").append(parsepath.m_path).append(parsepath.m_file)).c_str(), &FindFileData);
	if (hfind == INVALID_HANDLE_VALUE)
		throw std::runtime_error(MAKE_ERROR_MSG("FindFirstFile returned an Invalid Handle. " + getWinErrorMsg()));
	
	std::wstring filename;
	std::wstring completepath;
	do
	{
		if (lstrcmp(FindFileData.cFileName, L"..") == 0 || lstrcmp(FindFileData.cFileName, L".") == 0 || (FindFileData.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN) == FILE_ATTRIBUTE_HIDDEN)
			continue;
		else
		{
			filename = FindFileData.cFileName;
			completepath = parsepath.m_path;
			completepath += filename;
			if ((FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == FILE_ATTRIBUTE_DIRECTORY)
			{
				FolderVector::iterator i = sublist.begin();
				for ( ; i != sublist.end(); ++i)
				{
					if ((*i)->getEntry() == filename)
					{
						(*i)->appendPath(completepath);
						break;
					}
				}
				if (i == sublist.end())	// went throught them all
					sublist.push_back(createMenuFolderSpecial(filename, completepath, m_dynamic));
			}
			else
				itemlist.push_back(MenuItem::createMenuItem(completepath, std::wstring()));
		}
	}
	while (FindNextFile(hfind, &FindFileData));
	FindClose(hfind);
}

void MenuFolder::validatePath(StrVector& strs)
{
	if(strs.empty())
		return;

	std::wstring current = strs.back();
	strs.pop_back();

	std::wstring::size_type divider = current.find(L'|');
	while(divider != std::wstring::npos)
	{
		strs.push_back(current.substr(0, divider));
		current = current.substr(divider + 1);
		divider = current.find(L'|');
	}
	strs.push_back(current);	
}