/*
 * Copyright 2003, Nathan E. Moore
 * Parts of this file Copyright (C) 1997-2001 The LiteStep Development Team 
 * (see popupmaker.cpp in the popup2 source)
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "menusettings.h"
#include "menuitem.h"
#include "menubang.h"
#include "menumessage.h"
#include "menufolder.h"
#include "menushell.h"
#include "menutasks.h"
#include "globals.h"
#include <algorithm>
#include <windows.h>
#include <lsapi.h>
#include <cstring>
#include <stdexcept>
#include <shlwapi.h>

#define HOTLIST_HIDDEN "Litestep"

// prefix values
static const char* MENU = "Menu";
static const char* POPUP = "Popup";

// we cannot really say throw() as new throws a bad_alloc and both these use new
boost::shared_array<wchar_t> toUnicodeNoThrow(const char* astr);
boost::shared_array<wchar_t> expandWinEvars(boost::shared_array<wchar_t> str);
void nullDeleter(wchar_t*);

void MenuSettings::configure(BangVector& vector, BangCommandEx* f, HWND parent)
{
	const char* prefix = (GetRCBool("MenuUsePopup", TRUE) != 0) ? POPUP : MENU;
	settings(prefix);
	parseStepRC(prefix, vector, f, parent);
}

void MenuSettings::settings(const char* prefix)
{
	char* rclabel = new char[strlen(prefix) + 100];
	strcpy(rclabel, prefix);
	char* endprefix = rclabel + strlen(prefix);	

	// Popup2 settings that menu can use
	strcpy(endprefix, "NoTitles");
	MenuSub::header = GetRCBool(rclabel, FALSE) != 0;
	strcpy(endprefix, "ShowExtension");
	MenuItem::hasextensions = GetRCBool(rclabel, TRUE) != 0;

	// HotlistName is not prefixed with "Popup"
	char hlist[MAX_RC_STRING];
	char* hotlistlabel = "HotlistName";
	if (strcmp(POPUP, prefix) != 0)
		hotlistlabel = strcpy(endprefix, "HotListName");
	GetRCString(hotlistlabel, hlist, HOTLIST, MAX_RC_STRING);
	try
	{
		m_hotlist = toUnicode(hlist).get();
	}
	catch (std::exception& e)
	{
		LSLog(LOG_ERROR, MODULE_NAME, e.what());
		LSLog(LOG_ERROR, MODULE_NAME, "Due to above error using default value.");
		m_hotlist = W_HOTLIST;
	}

	// Menu settings that Popup2 doesn't use
	char horz[MAX_RC_STRING];
	char vert[MAX_RC_STRING];
	strcpy(endprefix, "HorzAnimation");
	GetRCString(rclabel, horz, "off", MAX_RC_STRING);
	strcpy(endprefix, "VertAnimation");
	GetRCString(rclabel, vert, "off", MAX_RC_STRING);
	MenuSub::setAnimation(horz, vert);

	strcpy(endprefix, "HorzPosition");
	GetRCString(rclabel, horz, "left", MAX_RC_STRING);
	strcpy(endprefix, "VertPosition");
	GetRCString(rclabel, vert, "top", MAX_RC_STRING);
	MenuSub::setPosition(horz, vert);

	strcpy(endprefix, "AutoColumn");
	MenuAuto::autocolumns = GetRCBool(rclabel, TRUE) != 0;
}

void MenuSettings::parseStepRC(const char* prefix, BangVector& vector, BangCommandEx* f, HWND parent)
{
	FILE *pStepRC = LCOpen(NULL);

	char szLineBuffer[MAX_LINE_LENGTH];
	char szFirst[MAX_PATH] = { 0 };
	char szSecond[MAX_PATH] = { 0 };
	char szThird[MAX_PATH] = { 0 };
	char szFourth[MAX_PATH] = { 0 };
	char szFifth[MAX_PATH] = { 0 };

	char *rgszTokens[] = { szFirst, szSecond, szThird, szFourth, szFifth };

	char rclabel[100];
	rclabel[0] = '*';
	strcpy(rclabel + 1, prefix);

	if(!LCReadNextConfig(pStepRC, rclabel, szLineBuffer, MAX_LINE_LENGTH))
		return; //Reads step.rc line starting with *Popup into szLineBuffer

	int nTokens = LCTokenize(szLineBuffer, rgszTokens, 5, NULL);
	//Tokenizes contents of szLineBuffer into up to 5 tokens,
	//stores them in rgszTokens with anything beyond 5 lost


	// multiple popup mode
	while(nTokens >= 2 && (strcmpi(szSecond, "!New") == 0 || strcmpi(szThird, "!New") == 0))
	{
		std::wstring caption;
		std::string bangcommand;
		char *rootcommand; // if after the bangcommand def there is another bang

		// set hotlist name and bang command acc to relevant arguments
		if(nTokens <= 3)
		{
			caption = m_hotlist;
			bangcommand = szThird;
			rootcommand = NULL;
		}
		else if (nTokens >= 4 && strcmpi(szSecond, "!New") == 0)
		{
			caption = m_hotlist;
			bangcommand = szThird;
			rootcommand = szFourth;
		}
		else if (nTokens >= 5 && strcmpi(szThird, "!New") == 0)
		{
			caption = toUnicodeNoThrow(szSecond).get();
			bangcommand = szFourth;
			rootcommand = szFifth;
		}
		else
		{
			caption = toUnicodeNoThrow(szSecond).get();
			bangcommand = szFourth;
			rootcommand = NULL;
		}

		// sets up a new PopupMenu with the spec'd bang command
		MenuSubPtr newpopup;

		if (rootcommand)
		{
			if (memicmp(rootcommand, "!PopupFolder:", strlen("!PopupFolder:")) == 0)
			{
				rootcommand += strlen("!PopupFolder:");
				newpopup = MenuFolder::createMenuFolder(caption, std::wstring(expandWinEvars(toUnicodeNoThrow(rootcommand)).get()), false);
			}
		}
		else
		{
			newpopup = MenuSub::createMenuSub(caption);
			parseFolder(rclabel, pStepRC, newpopup);
		}
		
		vector.push_back(BangableMenu::createBangableMenu(bangcommand, newpopup, parent, f));

		if(!LCReadNextConfig(pStepRC, rclabel, szLineBuffer, MAX_LINE_LENGTH))
			break;

		szFirst[0] = 0;
		szSecond[0] = 0;
		szThird[0] = 0;
		szFourth[0] = 0;
		szFifth[0] = 0;

		nTokens = LCTokenize(szLineBuffer, rgszTokens, 5, NULL);
	}

	LCClose(pStepRC);
}

void MenuSettings::parseFolder(const char* prefix, FILE *pStepRC, MenuSubPtr pop)
{
	while (true)
	{
		char szLineBuffer[MAX_LINE_LENGTH];

		if (!LCReadNextConfig(pStepRC, prefix, szLineBuffer, MAX_LINE_LENGTH))
		{
			break;
		}
		try
		{
			if (!parseLine(prefix, pStepRC, pop, szLineBuffer))
				break;
		}
		catch (std::exception& e)
		{
			LSLog(LOG_ERROR, MODULE_NAME, e.what());
		}
	}
}

bool MenuSettings::parseLine(const char* prefix, FILE *pStepRC, MenuSubPtr pop, char* pszLine)
{
	char szToken1[MAX_LINE_LENGTH], szToken2[MAX_LINE_LENGTH],
		 szToken3[MAX_LINE_LENGTH], szExtra[MAX_LINE_LENGTH];
	
    LPSTR lpszTokens[] = { szToken1, szToken2, szToken3 };

	szToken1[0] = szToken2[0] = szToken3[0] = szExtra[0] = '\0';

	int nTokens = LCTokenize(pszLine, lpszTokens, 3, szExtra);

	if (strnicmp(szToken2, ".icon=", 6) == 0)
	{
		char *rgszTokens[1] = { szToken3 };
		char szBuffer[MAX_LINE_LENGTH];

		strcpy(szToken2, szToken3);
		strcpy(szBuffer, szExtra);
		LCTokenize(szBuffer, rgszTokens, 1, szExtra);

		if (!szBuffer[0])
			nTokens--;
	}

	if (nTokens == 3)
	{
		if (!strcmpi(szToken3, "Folder"))
		{
			// Folder
			MenuSubPtr tempfolder = MenuSub::createMenuSub(toUnicodeNoThrow(szToken2).get());
			pop->addSubMenu(tempfolder, getColumn());
			parseFolder(prefix, pStepRC, tempfolder); 
		}
		else if (memicmp(szToken3, "!PopupFolder:", strlen("!PopupFolder:")) == 0)
		{
			// PopupFolder
			LPSTR pszTemp = szToken3 + strlen("!PopupFolder:");
			pop->addSubMenu(MenuFolder::createMenuFolder(toUnicodeNoThrow(szToken2).get(), expandWinEvars(toUnicodeNoThrow(pszTemp)).get(), false), getColumn());			
		}
		else if (memicmp(szToken3, "!DynamicFolder:", strlen("!DynamicFolder:")) == 0)
		{
			// PopupFolder
			LPSTR pszTemp = szToken3 + strlen("!DynamicFolder:");
			pop->addSubMenu(MenuFolder::createMenuFolder(toUnicodeNoThrow(szToken2).get(), expandWinEvars(toUnicodeNoThrow(pszTemp)).get(), true), getColumn());
		}
		else if (memicmp(szToken3, "!PopupDynamicFolder:", strlen("!PopupDynamicFolder:")) == 0)
		{
			// PopupFolder
			LPSTR pszTemp = szToken3 + strlen("!PopupDynamicFolder:");
			pop->addSubMenu(MenuFolder::createMenuFolder(toUnicodeNoThrow(szToken2).get(), expandWinEvars(toUnicodeNoThrow(pszTemp)).get(), true), getColumn());
		}
		else if(!strcmp(szToken2, "!RemoteAmp"))
		{
			MenuSubPtr sub;
			if (PathFileExistsA(szToken3) == TRUE)
				sub = createRemoteAmp(REMOTEAMP);
			else
				sub = createRemoteAmp(toUnicodeNoThrow(szToken3).get());
			pop->addSubMenu(sub, getColumn());
		}
		// if it is a special menu item that wants a bmp background
		//else if (szToken2[0] == '!')
		//{
			// This is the part that handles some of the popup2 specfic psuedobangs.
			// But None that i have impelmented;
		//}
		else
		{
			try
			{
				pop->addSubMenu(makeBangItem(szToken3, toUnicodeNoThrow(szToken2).get()), getColumn());
			}
			catch (...)
			{
				MenuEntryPtr entry;
				if (szToken3[0] == '!')
					entry = MenuBang::createMenuBang(toUnicodeNoThrow(szToken2).get(), szToken3, szExtra);
				else
					entry = MenuItem::createMenuItem(toUnicodeNoThrow(szToken2).get(), expandWinEvars(toUnicode(szToken3)).get(), toUnicode(szExtra).get());	
				pop->addItem(entry, getColumn());
			}
		}
	}
	else if (nTokens == 2)
	{
		if (strcmpi(szToken2, "~Folder") == 0 || strcmpi(szToken2, "~New") == 0)
			return false;
		else if (strcmpi(szToken2, "!Separator") == 0)
			pop->addSeparator(getColumn());
		else if (strcmpi(szToken2, "!Column") == 0)
			setColumn();
		else
			pop->addSubMenu(makeBangItem(szToken2), getColumn());
	}
	return true;;
}
MenuSubPtr MenuSettings::makeBangItem(const char* item, const std::wstring& title) const
{
	if (strcmpi(item, "!PopupTasks") == 0)
		return MenuTasks::createMenuTasks(title);
	else if (strcmpi(item, "!PopupControlPanel") == 0)
		return MenuShell::createMenuShell(title, CSIDL_CONTROLS);
	else if (strcmpi(item, "!PopupMyComputer") == 0)
		return MenuShell::createMenuShell(title, CSIDL_DRIVES);
	else if (strcmpi(item, "!PopupNetwork") == 0)
		return MenuShell::createMenuShell(title, CSIDL_NETWORK);
	else if (strcmpi(item, "!PopupPrinters") == 0)
		return MenuShell::createMenuShell(title, CSIDL_PRINTERS);
	else if (strcmpi(item, "!PopupRecentDocuments") == 0)
		return MenuShell::createMenuShell(title, CSIDL_RECENT);
	else if (strcmpi(item, "!PopupRecycleBin") == 0)
		return MenuShell::createMenuShell(title, CSIDL_BITBUCKET);
	else if (strcmpi(item, "!PopupTheDesktop") == 0)
		return MenuShell::createMenuShell(title, CSIDL_DESKTOP);
	else 
		throw std::runtime_error(MAKE_ERROR_MSG(std::string(item) + " is not a real item."));
}

MenuSubPtr MenuSettings::makeBangItem(const char* item) const
{
	if (strcmpi(item, "!PopupTasks") == 0)
		return MenuTasks::createMenuTasks();
	else if (strcmpi(item, "!PopupControlPanel") == 0)
		return MenuShell::createMenuShell(CSIDL_CONTROLS);
	else if (strcmpi(item, "!PopupMyComputer") == 0)
		return MenuShell::createMenuShell(CSIDL_DRIVES);
	else if (strcmpi(item, "!PopupNetwork") == 0)
		return MenuShell::createMenuShell(CSIDL_NETWORK);
	else if (strcmpi(item, "!PopupPrinters") == 0)
		return MenuShell::createMenuShell(CSIDL_PRINTERS);
	else if (strcmpi(item, "!PopupRecentDocuments") == 0)
		return MenuShell::createMenuShell(CSIDL_RECENT);
	else if (strcmpi(item, "!PopupRecycleBin") == 0)
		return MenuShell::createMenuShell(CSIDL_BITBUCKET);
	else if (strcmpi(item, "!PopupTheDesktop") == 0)
		return MenuShell::createMenuShell(CSIDL_DESKTOP);
	else 
		throw std::runtime_error(MAKE_ERROR_MSG(std::string(item) + " is not a real item."));
}


void nullDeleter(wchar_t*)
{	}

boost::shared_array<wchar_t> toUnicodeNoThrow(const char* astr)
{
	try
	{
		return toUnicode(astr);
	}
	catch (std::exception& e)
	{
		LSLog(LOG_ERROR, MODULE_NAME, e.what());
		return boost::shared_array<wchar_t>(L"ERROR!", nullDeleter);;
	}
}

boost::shared_array<wchar_t> expandWinEvars(boost::shared_array<wchar_t> str)
{
	DWORD needsize = ExpandEnvironmentStrings(str.get(), NULL, 0);
	boost::shared_array<wchar_t> newstr(new wchar_t[needsize]);
	DWORD good = ExpandEnvironmentStrings(str.get(), newstr.get(), needsize);
	if (!good)
		return str;
	else if (good <= needsize) // should aways be true, but it is better to be safe
		// TODO: change '|' to a cont value defined in menufolder.h
		std::replace(newstr.get(), newstr.get() + good, L';', L'|');
	return newstr;
}

MenuSubPtr createRemoteAmp(const std::wstring& label)
{
	MenuSubPtr remoteamp = MenuSub::createMenuSub(label);
	static const wchar_t* winampclass = L"Winamp v1.x";
	remoteamp->addItem(MenuMessage::createMenuMessage(std::wstring(L"Previous"), winampclass, NULL, WM_COMMAND, 40044, 0), false);
	remoteamp->addItem(MenuMessage::createMenuMessage(std::wstring(L"Play"), winampclass, NULL, WM_COMMAND, 40045, 0), false);
	remoteamp->addItem(MenuMessage::createMenuMessage(std::wstring(L"Pause"), winampclass, NULL, WM_COMMAND, 40046, 0), false);
	remoteamp->addItem(MenuMessage::createMenuMessage(std::wstring(L"Stop"), winampclass, NULL, WM_COMMAND, 40047, 0), false);
	remoteamp->addItem(MenuMessage::createMenuMessage(std::wstring(L"Next"), winampclass, NULL, WM_COMMAND, 40048, 0), false);
	return remoteamp;
}