/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "menushell.h"
#include "globals.h"
#include <shlobj.h>
#include <shlwapi.h>
#include <shtypes.h>
#include <algorithm>
#include <functional>
#include <lsapi.h>

// This is an optimization, as the util::ItemIdList comparison functions
// have to recreate a util::Desktop every time they are called this does not.
// All the constructors and operators are a temp measure, as std::sort requires a
// copiable binary predicated 
template<class lhs, class rhs>
class Sorter
	:public std::binary_function<lhs, rhs, bool>
{
private:
	util::InterfacePtr<IShellFolder> desk;
public:
	Sorter()
		:desk(util::getDesktop())
	{	}

	result_type operator()(first_argument_type& lhs, second_argument_type& rhs)
	{
		HRESULT hres = desk->CompareIDs(0, lhs->get(), rhs->get());
		return static_cast<short>(HRESULT_CODE(hres)) < 0;
	}
};
 
void MenuShell::parse()
{
	ShellVector folders;
	IdListVector items;

	parseFolder(folders, items);
	if (getEntry() == L"temp3")
		int x = 0-3+3;
	std::sort(folders.begin(), folders.end(), Sorter<MenuShellPtr, MenuShellPtr>());
	std::sort(items.begin(), items.end(), Sorter<MenuIdListPtr, MenuIdListPtr>());
	
	// This loop will merge the two vectors
	ShellVector::iterator f = folders.begin();
	IdListVector::iterator i = items.begin();
	Sorter<MenuShellPtr, MenuIdListPtr> comp;
	while(f != folders.end() && i != items.end())
	{
		if (comp(*f,*i))
		{
			addSubMenu(boost::static_pointer_cast<MenuSub>(*f));
            ++f;
		}
		else
		{
			addItem(boost::static_pointer_cast<MenuEntry>(*i));
			++i;
		}
	}

	if (f != folders.end())
	{
		for ( ; f != folders.end(); ++f)
			addSubMenu(boost::static_pointer_cast<MenuSub>(*f));
	}
	else if (i != items.end())
	{
		for ( ; i != items.end(); ++i)
			addItem(boost::static_pointer_cast<MenuEntry>(*i));
	}
}

void MenuShell::parseFolder(ShellVector &folderlist, IdListVector &itemlist)
{
	util::InterfacePtr<IShellFolder> folder;
	if (get()->mkid.cb == 0)	// Allow !PopupDesktop to work
		folder = util::getDesktop();
	else
	{
		IShellFolder* psf;
		if (FAILED(util::getDesktop()->BindToObject(get(), NULL, IID_IShellFolder, reinterpret_cast<void**>(&psf))))
			throw std::runtime_error(MAKE_ERROR_MSG("BindToObject Failed"));
		folder.reset(psf);
	}

	IEnumIDList* penumlist;
	if (FAILED(folder->EnumObjects(NULL, SHCONTF_NONFOLDERS | SHCONTF_FOLDERS , &penumlist)))
		throw std::runtime_error(MAKE_ERROR_MSG("Could not get IEnumIDList"));
	util::InterfacePtr<IEnumIDList> enumerator(penumlist);


	// declare the ItemIdList here to avoid having to recreate the enclosed Malloc
	util::ItemIdList currentitem;
	ITEMIDLIST* pidl;
	HRESULT hres;
	while (SUCCEEDED(hres = enumerator->Next(1, &pidl, NULL)) && hres != S_FALSE )
	{
		currentitem.reset(pidl);
		STRRET name;
		// Find Display Name
		if(SUCCEEDED(folder->GetDisplayNameOf(currentitem.get(), SHGDN_NORMAL, &name)))
		{
			wchar_t* strname;
			// Just Skip the item if we can't get a display name
			if (SUCCEEDED(StrRetToStr(&name, currentitem.get(), &strname)))
			{
				try
				{
					DWORD attrib = SFGAO_FOLDER | SFGAO_REMOVABLE;
					ITEMIDLIST* citemtemp = currentitem.get();
					// not sure why it needs the const_cast here, I fucking hate com
					if(SUCCEEDED(folder->GetAttributesOf(1, const_cast<const ITEMIDLIST**>(&citemtemp), &attrib)))
					{
						ItemIdList newidlist = concat(currentitem);
						if ((attrib & SFGAO_FOLDER) == SFGAO_FOLDER && (attrib & SFGAO_REMOVABLE) != SFGAO_REMOVABLE)
							folderlist.push_back(createMenuShellSpecial(strname, newidlist));
						else
							itemlist.push_back(MenuIdList::createMenuIdListSpecial(strname, newidlist));
					}
					CoTaskMemFree(strname);	// Could this use malloc as CoTaskMemFree uses the same this as IMalloc?
				}
				catch (std::exception& e)
				{
					LSLog(LOG_DEBUG, MODULE_NAME, e.what());
					CoTaskMemFree(strname);
				}
			}	
		}
	}
}