/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "MenuSub.h"
#include "globals.h"
#include <lsapi.h>
#include <utility>
#include <algorithm>
#include <stdexcept>
#include <cassert>

bool MenuSub::header = false;
long MenuSub::animation = TPM_NOANIMATION;
long MenuSub::position = TPM_LEFTALIGN | TPM_TOPALIGN;

MenuSub::MenuSub(const std::wstring &txt)
	:MenuEntry(txt),
	m_parent(NULL)
{	
	if (header)
	{
		addDeadItem(getEntry());
		SetMenuDefaultItem(m_menu, 0, TRUE);	// Make the Header Bold
		addSeparator();
	}
}

void MenuSub::addItem(const MenuEntryPtr &item, bool column)
{ 
	m_menu.appendItem(MF_ENABLED | MF_STRING | (column ? MF_MENUBARBREAK : 0), item->getId(), item->getEntry().c_str());
	try
	{
		// assign the result to a var to insure that insert actually gets called in a NDEBUG build
		bool fail = m_itemmap.insert(ItemMap::value_type(item->getId(), item)).second;
		assert(fail);
	}
	catch (...)
	{
		m_menu.removeItem(item->getId());
		throw;
	}
}

void MenuSub::addSeparator(bool column)
{
	m_menu.appendItem(MF_SEPARATOR | (column ? MF_MENUBARBREAK : 0), NULL_ID, NULL);
}

void MenuSub::addSubMenu(MenuSubPtr &sub, bool column)
{
	m_menu.appendItem(MF_ENABLED | MF_POPUP | MF_STRING | (column ? MF_MENUBARBREAK : 0),
		reinterpret_cast<UINT_PTR>(sub->m_menu.get()), sub->getEntry().c_str());
	try
	{
		bool fail = m_submap.insert(SubMap::value_type(sub->m_menu, sub)).second;
		assert(fail);
		sub->m_parent = &m_menu;
	}
	catch (...)
	{
		m_menu.removeItem(reinterpret_cast<UINT_PTR>(sub->m_menu.get()));
		throw;
	}
}

void MenuSub::addDeadItem(const std::wstring &txt, bool column)
{
	m_menu.appendItem(MF_ENABLED | MF_STRING | MF_DISABLED | (column ? MF_MENUBARBREAK : 0), NULL_ID, txt.c_str());
}

void MenuSub::displayMenu(int x, int y, HWND hwnd)
{
	if (!TrackPopupMenuEx(m_menu, TPM_LEFTBUTTON | animation | position, x, y, hwnd, NULL))
		throw std::runtime_error(MAKE_ERROR_MSG("TrackPopupMenuEx Failed. " + getWinErrorMsg()));
}

MenuEntryPtr MenuSub::getItem(unsigned int searchid) const
{
	// item is in this folder
	ItemMap::const_iterator itemiter = m_itemmap.find(searchid);
	if (itemiter != m_itemmap.end())
		return (itemiter->second);

	// item is in a subfolder;
	MenuEntryPtr item;
	for(SubMap::const_iterator i = m_submap.begin(); i != m_submap.end(); ++i)
	{
		item = (i->second)->getItem(searchid);
		if (item)
			return item;
	}

	// couldn't find it, return NULL
	item.reset();
	return item;
}

MenuSubPtr MenuSub::getSub(HMENU searchhmenu) const
{
	SubMap::const_iterator subiter = m_submap.find(searchhmenu);
	if (subiter != m_submap.end())
		return subiter->second;

	MenuSubPtr sub;
	for (SubMap::const_iterator i = m_submap.begin(); i != m_submap.end(); ++i)
	{
		sub = (i->second)->getSub(searchhmenu);
		if (sub)
			return sub;
	}
	sub.reset();
	return sub;
}

void MenuSub::endDisplay()
{
	// Should I remove endDisplay() since all it does it call postDisplay()?
	postDisplay();
}

void MenuSub::setEntry(std::wstring &newentry)
{
	MenuEntry::setEntry(newentry);
	
	if (header)
		m_menu.modifyItem(0, MF_BYPOSITION, NULL_ID, getEntry().c_str());
}

void MenuSub::postDisplay()
{
	for (SubMap::iterator i = m_submap.begin(); i != m_submap.end(); ++i)
		i->second->postDisplay();	
}

void MenuSub::setAnimation(const char* horzmode, const char* vertmode)
{
	long horz = 0L;
	long vert = 0L;
    if (lstrcmpiA(horzmode, "off") == 0)
		horz = TPM_NOANIMATION;
	else if (lstrcmpiA(horzmode, "right") == 0)
		horz = TPM_HORNEGANIMATION;
	else if (lstrcmpiA(horzmode, "left") == 0)
		horz = TPM_HORPOSANIMATION;

	if (lstrcmpiA(vertmode, "off") == 0)
		vert = TPM_NOANIMATION;
	else if (lstrcmpiA(vertmode, "up") == 0)
		vert = TPM_VERNEGANIMATION;
	else if (lstrcmpiA(vertmode, "down") == 0)
		vert = TPM_VERPOSANIMATION;

	animation = 0;
	if (horz == TPM_NOANIMATION && vert == TPM_NOANIMATION)
		animation = TPM_NOANIMATION;
	else if (horz == TPM_NOANIMATION)
		animation = vert;
	else if (vert == TPM_NOANIMATION)
		animation = horz;
	else
		animation = horz | vert;
}

void MenuSub::setPosition(const char* horzmode, const char* vertmode)
{
	if (lstrcmpiA(horzmode, "left") == 0)
		position = TPM_LEFTALIGN;
	else if (lstrcmpiA(horzmode, "center") == 0)
		position = TPM_CENTERALIGN;
	else if (lstrcmpiA(horzmode, "right") == 0)
		position = TPM_RIGHTALIGN;

	if (lstrcmpiA(vertmode, "top") == 0)
		position |= TPM_TOPALIGN;
	else if (lstrcmpiA(vertmode, "center") == 0)
		position |= TPM_VCENTERALIGN;
	else if (lstrcmpiA(vertmode, "bottom") == 0)
		position |= TPM_BOTTOMALIGN;
}