/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#pragma once
#include <windows.h>
#include <boost/noncopyable.hpp>
#include <stdexcept>
#include "globals.h"

namespace util
{
	class SmartHmenu
		:public boost::noncopyable
	{
	private:
		HMENU m_hmenu;

	public:
		SmartHmenu() throw(...)
			:m_hmenu(CreatePopupMenu())
		{
			if (!m_hmenu)
				throw std::runtime_error(MAKE_ERROR_MSG("CreatePopupMenu Failed. " + getWinErrorMsg()));
		}
		
		SmartHmenu(HMENU hmenu)
			:m_hmenu(hmenu)
		{	}

		~SmartHmenu()
		{
			if (m_hmenu)
			{
				DestroyMenu(m_hmenu);
				m_hmenu = NULL;
			}
		}

		HMENU release()
		{
			HMENU temp = m_hmenu;
			m_hmenu = NULL;
			return temp;
		}

		operator const HMENU() const
		{
			return m_hmenu;
		}

		HMENU get()
		{
			return m_hmenu;
		}

		void appendItem(UINT flags, UINT_PTR id, LPCWSTR entry)
		{
			if (!AppendMenu(m_hmenu, flags, id, entry))
				throw std::runtime_error(MAKE_ERROR_MSG("AppendMenu Failed. " + getWinErrorMsg()));
		}

		void removeItem(UINT pos, UINT flags = MF_BYCOMMAND)
		{
			if (!RemoveMenu(m_hmenu, pos, flags))
				throw std::runtime_error(MAKE_ERROR_MSG("RemoveMenu Failed. " + getWinErrorMsg()));
		}

		void modifyItem(UINT pos, UINT flags, UINT_PTR id, LPCWSTR entry)
		{
			if (!ModifyMenu(m_hmenu, pos, flags, id, entry))
				throw std::runtime_error(MAKE_ERROR_MSG("ModifyMenu Failed. " + getWinErrorMsg()));
		}

		void deleteItem(UINT pos, UINT flags = MF_BYCOMMAND)
		{
			if (!DeleteMenu(m_hmenu, pos, flags))
				throw std::runtime_error(MAKE_ERROR_MSG("DeleteMenu Failed. " + getWinErrorMsg()));
		}

		void getInfo(LPMENUINFO minfo)
		{
			if (!GetMenuInfo(m_hmenu, minfo))
				throw std::runtime_error(MAKE_ERROR_MSG("GetMenuInfo Failed. " + getWinErrorMsg()));
		}

		void setInfo(LPCMENUINFO minfo)
		{
			if (!SetMenuInfo(m_hmenu, minfo))
				throw std::runtime_error(MAKE_ERROR_MSG("SetMenuInfo Failed. " + getWinErrorMsg()));
		}

		void getItemInfo(UINT pos, UINT flags, LPMENUITEMINFO mii)
		{
			if (!GetMenuItemInfo(m_hmenu, pos, flags, mii))
				throw std::runtime_error(MAKE_ERROR_MSG("GetMenuItemInfo Failed. " + getWinErrorMsg()));
		}

		void setItemInfo(UINT pos, UINT flags, LPMENUITEMINFO mii)
		{
			if (!SetMenuItemInfo(m_hmenu, pos, flags, mii))
				throw std::runtime_error(MAKE_ERROR_MSG("SetMenuItemInfo Failed. " + getWinErrorMsg()));			
		}
		
		void setDefaultItem(UINT pos, UINT flags = MF_BYCOMMAND)
		{
			if (!SetMenuDefaultItem(m_hmenu, pos, flags))
				throw std::runtime_error(MAKE_ERROR_MSG("SetMenuDefaultItem Failed. " + getWinErrorMsg()));
		}

		int size()
		{
			int size = GetMenuItemCount(m_hmenu);
			if (size == -1)
				throw std::runtime_error(MAKE_ERROR_MSG("GetMenuItemCount Failed. " + getWinErrorMsg()));
			return size;
		}

		RECT getItemRect(HWND wnd, UINT pos)
		{
			RECT rc;
			if (!GetMenuItemRect(wnd, m_hmenu, pos, &rc))
				throw std::runtime_error(MAKE_ERROR_MSG("GetMenuItemRect Failed. " + getWinErrorMsg()));
			return rc;

		}
	};
}