/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#pragma once
#include "globals.h"
#include <stdexcept>
#include <new>
#include <shtypes.h>
#include <shlobj.h>
#include <stddef.h>	// for size_t

namespace util
{
	template<class T>
	class InterfacePtr
	{
	protected:
		T* m_ptr;

		void free() throw()
		{
			if (m_ptr)
				m_ptr->Release();
		}

	public:
		explicit InterfacePtr(T* ptr = 0) throw()
			:m_ptr(ptr)
		{	}

		InterfacePtr(InterfacePtr<T>& ifaceptr) throw()
			:m_ptr(ifaceptr.m_ptr)
		{
			if (m_ptr)
				m_ptr->AddRef();
		}

		InterfacePtr<T> & operator=(InterfacePtr<T> ifaceptr) throw()
		{
			reset(ifaceptr.m_ptr);
			if (m_ptr)
				m_ptr->AddRef();
			return *this;
		}

		virtual ~InterfacePtr() throw()
		{
			free();
		}
		
		T* get() const throw()
		{
			return m_ptr;
		}

		T* operator->() const throw()
		{
			return get();
		}

		// named remove instead of the standard release to prevent confusion
		T* remove() throw()
		{
			T* temp = m_ptr;
			m_ptr = NULL;
			return temp;
		}

		void reset(T* ptr = 0) throw()
		{
			if (ptr != m_ptr)
				free();
			m_ptr = ptr;
		}

		operator bool () const throw()
		{
			return m_ptr != 0;
		}
	};

	// The Point of this class is to allow for checking of the return values from
	// the IMalloc allocator, it has simalar behavior to the use of new/delete
	class Malloc
		:public InterfacePtr<IMalloc>
	{
	public:
		Malloc()
		{
			if(FAILED(SHGetMalloc(&m_ptr)))
				throw std::runtime_error(MAKE_ERROR_MSG("Could not get IMalloc"));
		}

		void* alloc(size_t count) const throw(...)
		{
			void* rval = m_ptr->Alloc(count);
			if (!rval)
				throw std::bad_alloc(MAKE_ERROR_MSG("Alloc Failed").c_str());
			return rval;
		}

		void free(void* ptr) const throw()
		{
			m_ptr->Free(ptr);
			ptr = NULL;
		}
	};

	// std::auto_ptr clone using IMalloc instead of new/delete
	template<class T>
	class IMallocPtr
	{
	private:
		void free() throw()
		{
			if (m_ptr)
			{
				m_imalloc.free(m_ptr);
				m_ptr = NULL;
			}
		}
		
	protected:
		T* m_ptr;
		Malloc m_imalloc;

	public:
		explicit IMallocPtr(T* ptr = 0) throw()
			:m_ptr(ptr)
		{	}

		IMallocPtr(IMallocPtr<T>& imallocptr) throw()
			:m_ptr(imallocptr.release())
		{	}

		IMallocPtr<T> & operator=(IMallocPtr<T> imallocptr) throw()
		{
			reset(imallocptr.release());
			return *this;
		}

		virtual ~IMallocPtr() throw()
		{
			free();
		}
		
		T* get() const throw()
		{
			return (m_ptr);
		}

		T* operator->() const throw()
		{
			return (get());
		}

		T* release() throw()
		{
			T* temp = m_ptr;
			m_ptr = NULL;
			return (temp);
		}

		void reset(T* ptr = 0) throw()
		{
			if (ptr != m_ptr)
				free();
			m_ptr = ptr;
		}

		operator bool () const throw()
		{
			return m_ptr != 0;
		}
	};
}