/*
 * Copyright 2003, Nathan E. Moore
 *
 * This file is part of Menu.
 *
 * Menu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Menu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Menu; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "idList.h"
#include "smarthmenu.h"
#include "globals.h"
#include <windows.h>
#include <shlwapi.h>
#include <boost/scoped_array.hpp>
#include <cassert>

namespace util
{
	ItemIdList createVirtualFolder(int csidl)
	{
		ITEMIDLIST* pidl;
		if(FAILED(SHGetFolderLocation(NULL, csidl, NULL, NULL, &pidl)))
			throw std::runtime_error(MAKE_ERROR_MSG("SHGetFolderLocation Failed"));
		return ItemIdList(pidl);
	}

	InterfacePtr<IShellFolder> getDesktop()
	{
		IShellFolder* psf;
		if(FAILED(SHGetDesktopFolder(&psf)))
			throw std::runtime_error(MAKE_ERROR_MSG("SHGetDesktopFolder failed"));
		return InterfacePtr<IShellFolder>(psf);
	}
	//
	// ItemIdList
	//
	// Determines the size in bytes of the give ITEMIDLIST*
	size_t ItemIdList::sizeOf(ITEMIDLIST* pidl) const
	{
		// a null pointer has no size
		assert(pidl);
		
		size_t size = 2;	// start at 2 for the null
		ITEMIDLIST* current = pidl;
		while (current->mkid.cb)
		{
			size += current->mkid.cb;
			current = next(current);
		}
		return size;
	}

	// copies pidl to the return value, the rvalue must be freed by IMalloc
	ITEMIDLIST* ItemIdList::copy(ITEMIDLIST* pidl) const
	{
		assert(pidl);

		size_t size = sizeOf(pidl);
		ITEMIDLIST* rval = static_cast<ITEMIDLIST*>(m_imalloc.alloc(size));
		CopyMemory(rval, pidl, size);
		return rval;
	}

	// appends idlist.m_ptr to the end of this.m_ptr
	// for making multilevel pidls
	ItemIdList & ItemIdList::append(const ItemIdList &idlist)
	{
		if (idlist.m_ptr == NULL)
			return *this;
		else if (m_ptr == NULL)
		{
			reset(copy(idlist.m_ptr));
			return *this;
		}
		else
		{
			size_t sizethis = sizeOf(m_ptr);
			size_t sizeidlist = sizeOf(idlist.m_ptr);
			size_t sizenew = sizethis + sizeidlist - 2;

			ITEMIDLIST* pidlnew = static_cast<ITEMIDLIST*>(m_imalloc.alloc(sizenew));
			CopyMemory(pidlnew, m_ptr, sizethis - 2);
			unsigned char* end1 = reinterpret_cast<unsigned char*>(pidlnew) + sizethis - 2;
			CopyMemory(end1, idlist.m_ptr, sizeidlist);
			reset(pidlnew);
			return *this;
		}
	}

	// returns a pointer to the last element of m_list
	ITEMIDLIST* ItemIdList::file() const
	{
		ITEMIDLIST* file = m_ptr;
		ITEMIDLIST* end = file;
		while (end->mkid.cb)
		{
			file = end;
			end = next(end);
		}
		return file;
	}
    
	// returns a new smartptr holding all but the last element in m_ptr
	ItemIdList ItemIdList::getContainingElement() const
	{
		size_t size = sizeOf(m_ptr) - sizeOf(file()) + 2;
		ITEMIDLIST* newidlist = static_cast<ITEMIDLIST*>(m_imalloc.alloc(size));
		ZeroMemory(newidlist, size);
		CopyMemory(newidlist, m_ptr, size - 2);
		//*(reinterpret_cast<USHORT*>((reinterpret_cast<char*>(newidlist) + size))) = 0;
		return ItemIdList(newidlist);
	}

	// executes the file represted by the ItemIdList
	void ItemIdList::execute(HWND hwnd) throw(...)
	{
		wchar_t folder[MAX_PATH];
		SHGetPathFromIDList(getContainingElement().m_ptr, folder);

		SHELLEXECUTEINFO info;
		ZeroMemory(&info, sizeof(info));
		info.cbSize = sizeof(info);
		info.fMask = SEE_MASK_IDLIST | SEE_MASK_FLAG_NO_UI | SEE_MASK_INVOKEIDLIST;
		info.lpIDList = static_cast<void*>(m_ptr);
		info.hwnd = hwnd;
		info.nShow = SW_SHOW;
		info.lpDirectory = folder;

		if (!ShellExecuteEx(&info))
			throw std::runtime_error(MAKE_ERROR_MSG("ShellExecute Failed" + getWinErrorMsg()));
	}

	std::wstring ItemIdList::getDisplayName() const throw(...)
	{
		IShellFolder* psf;
		const ITEMIDLIST* singlelevel;
		if (FAILED(SHBindToParent(m_ptr, IID_IShellFolder , (void**) &psf, &singlelevel)))
			throw std::runtime_error(MAKE_ERROR_MSG("Bind to Parent Failed"));
		InterfacePtr<IShellFolder> parent(psf);
		STRRET name;
		HRESULT hres = parent->GetDisplayNameOf(singlelevel, SHGDN_NORMAL, &name);
		if (FAILED(hres))
			throw std::runtime_error(MAKE_ERROR_MSG("Could not get display name"));
		wchar_t* strname;
		if (FAILED(StrRetToStr(&name, singlelevel, &strname)))
			throw std::runtime_error(MAKE_ERROR_MSG("StrRetToStr Failed"));
		try
		{
			std::wstring rvalue(strname);
			CoTaskMemFree(strname);
			return rvalue;
		}
		catch (...)
		{
			CoTaskMemFree(strname);
			throw;
		}
	}

	//
	// ContextMenu
	//

	IContextMenu2* ContextMenu::pcm2 = NULL;

	bool ContextMenu::popupContextMenu(const ItemIdList &idlist, HWND hwnd)
	{
		using std::runtime_error;

		IShellFolder* isf;
		ITEMIDLIST* filelist;
		if (FAILED(SHBindToParent(idlist.get(), IID_IShellFolder, reinterpret_cast<void**>(&isf), const_cast<const ITEMIDLIST**>(&filelist))))
			throw runtime_error(MAKE_ERROR_MSG("SHBindToParent Failed"));
		InterfacePtr<IShellFolder> folder(isf);
		IContextMenu* icm;
		if (FAILED(folder->GetUIObjectOf(hwnd, 1, const_cast<const ITEMIDLIST**>(&filelist), IID_IContextMenu, 0, reinterpret_cast<void**>(&icm))))
			throw runtime_error(MAKE_ERROR_MSG("GetUIObjectOf Failed"));
		InterfacePtr<IContextMenu> cmenu(icm);

		SmartHmenu contextmenu;
		if (FAILED(cmenu->QueryContextMenu(contextmenu, 0, 1, 0x7FFF, CMF_NORMAL)))
			throw runtime_error(MAKE_ERROR_MSG("QueryContextMenu Failed"));

		// No way for this to get lost before it is released, as nothing between this QueryInterface call
		// the SAFERELEASE(pcm2) below will throw any exceptions as it is all C or COM calls
		LONG proc;
		if (SUCCEEDED(cmenu->QueryInterface(IID_IContextMenu2, reinterpret_cast<void**>(&pcm2))))
		{
			// Subclassing the window is a very poor solution for this however,
			// I have yet to come up with something better when I do it will change
			proc = GetWindowLong(hwnd, GWL_WNDPROC);
			SetWindowLong(hwnd, GWL_WNDPROC, reinterpret_cast<LONG>(wndProc));
		}

		POINT pt;	
		GetCursorPos(&pt);
		int cmd = TrackPopupMenuEx(contextmenu, TPM_RECURSE | TPM_LEFTALIGN | TPM_TOPALIGN | TPM_LEFTBUTTON | TPM_RETURNCMD, pt.x, pt.y, hwnd, NULL);

		if (pcm2)
		{
			pcm2->Release();
			pcm2 = NULL;
			SetWindowLong(hwnd, GWL_WNDPROC, proc);
		}

		if (cmd != 0) 
		{
			CMINVOKECOMMANDINFO cmdinfo;
			cmdinfo.cbSize = sizeof(cmdinfo);
			cmdinfo.fMask = CMIC_MASK_FLAG_NO_UI;
			cmdinfo.hwnd = hwnd;
			// This calls a ASCII version of MAKEINTRESOURCE as the value passes is not a c string, but rathat a
			// fucked up pointer value including the menu id, so the A version differs from the W version only by
			// what it casts to
			cmdinfo.lpVerb = MAKEINTRESOURCEA(cmd - 1);
			cmdinfo.lpParameters = NULL;
			cmdinfo.lpDirectory = NULL;
			cmdinfo.nShow = SW_SHOW;

			if(FAILED(cmenu->InvokeCommand((CMINVOKECOMMANDINFO*)&cmdinfo)))
				throw runtime_error(MAKE_ERROR_MSG("InvokeCommand Failed"));
		}
		return (cmd != 0);
	}

	bool ContextMenu::popupContextMenu(const std::wstring &filesysobj, HWND hwnd)
	{
		// measure the size of the FullPath
		//DWORD pathsize = GetFullPathName(filesysobj.c_str(), 0, NULL, NULL);
		//boost::scoped_array<wchar_t> fullpath(new wchar_t[pathsize]);
		//GetFullPathName(filesysobj.c_str(), pathsize, fullpath.get(), NULL);

		boost::scoped_array<wchar_t> fullpath(new wchar_t[filesysobj.size() + 1]);
		filesysobj.copy(fullpath.get(), filesysobj.size());
		*(fullpath.get() + filesysobj.size()) = L'\0';

		ITEMIDLIST* pidl;
		if (SUCCEEDED(getDesktop()->ParseDisplayName(hwnd, NULL, fullpath.get(), NULL, &pidl, NULL)))
		{
			ItemIdList idlist(pidl);
			return popupContextMenu(idlist, hwnd);		
		}
		return false;
	}

	LRESULT CALLBACK ContextMenu::wndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
	{
		// TODO: should this pass any unspecified messages to the normal wndproc?
		switch (msg)
		{
		case WM_INITMENUPOPUP:
		case WM_DRAWITEM:
		case WM_MENUCHAR:
		case WM_MEASUREITEM:
			if(pcm2 != NULL)
				pcm2->HandleMenuMsg(msg, wparam, lparam);
			break;
		default:
			break;
		}
		return DefWindowProc(hwnd, msg, wparam, lparam);
	}
}